/*
 *  TopViewPanel  -- A panel that draws a 2D view (planform) of the wing geometry.
 *
 *  Copyright (C) 2006-2018, Joseph A. Huwaldt
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/gpl.html
 */

import jahuwaldt.js.ParamTools.AutoParam;
import jahuwaldt.tools.math.MathTools;
import java.awt.*;
import java.awt.geom.AffineTransform;
import java.awt.geom.GeneralPath;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.text.MessageFormat;
import java.util.Observable;
import java.util.Observer;
import java.util.ResourceBundle;
import javax.measure.unit.SI;
import javax.swing.JPanel;

/**
 * A panel component that draws a top view of the wing geometry.
 *
 * <p> Modified by: Joseph A. Huwaldt </p>
 *
 * @author Joseph A. Huwaldt, Date: January 15, 2006
 * @version December 16, 2023
 */
public class TopViewPanel extends JPanel implements Observer {

    //  Default values for the key parameters.
    private static final float DEFAULT_CR = 1;			//  meters
    private static final float DEFAULT_CT = 1;
    private static final float DEFAULT_SPAN = 4;
    private static final float DEFAULT_SWEEPLE = 0;

    //  The number of pixels or dots per meter (if drawn full scale) (dots/in*in/ft*ft/m).
    private static final float DPM = 72*12/0.3048F;

    //  The resource bundle for this panel's resources.
    private final ResourceBundle resBundle;
    
    //  The current drawing scale.
    private int modelScaleDenom = 100;
    private float scale = DPM/modelScaleDenom;

    //  The pens we are using to draw with.
    //private Stroke thinLine = new BasicStroke(0.00035F*modelScaleDenom);
    private Stroke thickLine = new BasicStroke(0.0007F*modelScaleDenom);
    private final float[] clDash = {0.038F*modelScaleDenom, 0.0032F*modelScaleDenom,
        0.0068F*modelScaleDenom, 0.0032F*modelScaleDenom};
    private Stroke centerLine = new BasicStroke(0.00035F*modelScaleDenom, BasicStroke.CAP_ROUND,
            BasicStroke.JOIN_BEVEL, 1, clDash, 0);

    //  The transformation to apply to the geometry before it is rendered.
    private AffineTransform at = AffineTransform.getScaleInstance(scale, scale);

    //  The geometry we are drawing.
    private final AutoParam crParam;		//  Root chord
    private final AutoParam ctParam;		//  Tip chord
    private final AutoParam bParam;         //  Span length
    private final AutoParam SweepLEParam;   //  Leading edge sweep angle.

    /**
     * The constructor for this class.
     *
     * @param resBundle The resource bundle for this panels resources.
     * @param cr Root chord length
     * @param ct Tip chord length
     * @param b Wing span length
     * @param SweepLE Leading edge sweep angle
     */
    @SuppressWarnings({"LeakingThisInConstructor", "OverridableMethodCallInConstructor"})
    public TopViewPanel(ResourceBundle resBundle, AutoParam cr, AutoParam ct, AutoParam b, AutoParam SweepLE) {
        this.resBundle = resBundle;
        this.crParam = cr;
        this.ctParam = ct;
        this.bParam = b;
        this.SweepLEParam = SweepLE;

        //  Link this object an observer of these parameters.
        cr.addObserver(this);
        ct.addObserver(this);
        b.addObserver(this);
        SweepLE.addObserver(this);

        this.setBackground(Color.white);
    }

    /**
     * This method sets the model drawing scale.
     *
     * @param denom The denominator of the scale factor (e.g.: 100 for 1/100
     * scale).
     */
    private void setScale(int denom) {

        modelScaleDenom = denom;
        scale = DPM/modelScaleDenom;
        //thinLine = new BasicStroke(0.00035F*denom);
        thickLine = new BasicStroke(0.0007F*denom);
        clDash[0] = 0.0038F*denom;
        clDash[1] = 0.0032F*denom;
        clDash[2] = 0.0068F*denom;
        clDash[3] = 0.0032F*denom;
        centerLine = new BasicStroke(0.00035F*denom, BasicStroke.CAP_ROUND,
                BasicStroke.JOIN_BEVEL, 1, clDash, 0);

        at = AffineTransform.getScaleInstance(scale, scale);
    }

    /**
     * Method that returns the denominator of the current model drawing scale.
     * For example, if the current scale is 1/200, then 200 is returned.
     * 
     * @return The denominator of the current drawing scale.
     */
    public int getScale() {
        return modelScaleDenom;
    }

    /**
     * Method that automatically chooses an appropriate scale for drawing the
     * geometry.
     */
    public void autoScale() {
        Dimension size = this.getSize();

        Point2D.Float ul = new Point2D.Float(getMinX(), getMinY());
        Point2D.Float lr = new Point2D.Float(getMaxX(), getMaxY());
        Point2D.Float dest = new Point2D.Float();

        boolean done = false;
        while (!done) {
            //  Zoom in until a point falls outside the view area.
            doZoomIn();
            AffineTransform transform = getTransform(size);

            //  Test to see if we zoomed to far.
            transform.transform(ul, dest);
            if (dest.x<0||dest.x>size.width||dest.y<0||dest.y>size.height)
                done = true;

            else {
                transform.transform(lr, dest);
                if (dest.x<0||dest.x>size.width||dest.y<0||dest.y>size.height)
                    done = true;
            }
        }

        //  Now zoom out until everything is contained in the panel.
        done = false;
        while (!done) {
            doZoomOut();
            AffineTransform transform = getTransform(size);
            done = true;

            //  Test to see if we have zoomed far enough.
            transform.transform(ul, dest);
            if (dest.x<0||dest.x>size.width||dest.y<0||dest.y>size.height)
                done = false;

            else {
                transform.transform(lr, dest);
                if (dest.x<0||dest.x>size.width||dest.y<0||dest.y>size.height)
                    done = false;
            }
        }

    }

    /**
     * Zoom in to the next highest acceptable scale factor.
     */
    private void doZoomIn() {
        int denom = getScale();

        int p10 = (int)(Math.log10(denom)+.1);
        int value = (int)((double)denom/MathTools.pow10(p10));
        switch (value) {
            case 1:
                value = 5;
                --p10;
                break;
            case 2:
                value = 1;
                break;
            default:
                value = 2;
                break;
        }

        denom = (int)(value*MathTools.pow10(p10));
        setScale(denom);
    }

    /**
     * Zoom out to the next lowest acceptable scale factor.
     */
    private void doZoomOut() {

        int denom = getScale();

        int p10 = (int)(Math.log10(denom)+.1);
        int value = (int)((double)denom/MathTools.pow10(p10));
        switch (value) {
            case 1:
                value = 2;
                break;
            case 2:
                value = 5;
                break;
            default:
                value = 1;
                ++p10;
                break;
        }

        denom = (int)(value*MathTools.pow10(p10));
        setScale(denom);
    }

    /**
     * Called automatically by Swing system to paint this component.
     * 
     * @param gc The graphics context into which the component is painted.
     */
    @Override
    public void paintComponent(Graphics gc) {

        // Paint the background
        super.paintComponent(gc);

        //  Scale the geometry to fit.
        autoScale();

        Graphics2D g2 = (Graphics2D)gc;
        g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);

        //  Draw a string describing the view.
        Dimension size = this.getSize();
        drawViewName(g2, size);

        //  Draw the scale factor.
        drawScaleFactor(g2, size);

        //  Save states that will need to be restored later.
        Stroke saveStroke = g2.getStroke();
        AffineTransform saveXform = g2.getTransform();

        //  Set up drawing transformations.
        AffineTransform transform = getTransform(size);
        g2.transform(transform);

        //  Render the geometry.
        renderWingGeometry(g2);

        //  Draw plane of symmetry line.
        g2.setStroke(centerLine);
        g2.draw(createSymmetryLine());

        //  Be kind, rewind.
        g2.setTransform(saveXform);
        g2.setStroke(saveStroke);
    }

    /**
     * Method that returns an appropriate transformation for this view panel.
     * Typically, this involves a flipping/rotating of the axes and a
     * translation to center the model in the panel.
     */
    private AffineTransform getTransform(Dimension panelSize) {

        //  Get the bounds of our geometry.
        float minSX = getMinX()*scale;
        float maxSX = getMaxX()*scale;
        float minSY = getMinY()*scale;
        float maxSY = getMaxY()*scale;

        //  Create a translation that will center the model in
        //  this panel.
        AffineTransform transform = new AffineTransform();
        transform.translate(panelSize.width/2-(maxSX+minSX)/2, panelSize.height/2-(maxSY+minSY)/2);
        transform.concatenate(at);

        return transform;
    }

    /**
     * Method that returns a line representing the plane of symmetry for this
     * view.
     */
    private Line2D createSymmetryLine() {
        float vlength = getMaxY()-getMinY();
        Line2D line = new Line2D.Float(0, -0.25F*vlength, 0, 1.25F*vlength);
        return line;
    }

    /**
     * Method that returns the name of this view. This method returns "Wing Planform".
     */
    private String getViewName() {
        return resBundle.getString("topViewName");
    }

    /**
     * Method that renders the model geometry as a wireframe.
     */
    private void renderWingGeometry(Graphics2D gc) {

        gc.setStroke(thickLine);

        double b = getSpanLength();
        float bo2 = (float)(b/2);
        float xt = (float)(bo2*Math.tan(getSweepLE()));

        //  Draw right-hand side.
        GeneralPath polygon = new GeneralPath(GeneralPath.WIND_EVEN_ODD, 5);

        polygon.moveTo(0, 0);
        polygon.lineTo(bo2, xt);
        polygon.lineTo(bo2, xt+getCtLength());
        polygon.lineTo(0, getCrLength());
        polygon.closePath();

        gc.draw(polygon);

        //  Draw left-hand side.
        polygon = new GeneralPath(GeneralPath.WIND_EVEN_ODD, 5);

        polygon.moveTo(0, 0);
        polygon.lineTo(-bo2, xt);
        polygon.lineTo(-bo2, xt+getCtLength());
        polygon.lineTo(0, getCrLength());
        polygon.closePath();

        gc.draw(polygon);
    }

    /**
     * Method that draws the view name to the bottom of the view.
     */
    private void drawViewName(Graphics2D g2, Dimension size) {

        int w = size.width;
        int h = size.height;
        String name = getViewName();

        FontMetrics metrics = g2.getFontMetrics();
        int width = metrics.stringWidth(name);

        g2.drawString(name, w/2-width/2, h-5);

    }

    /**
     * Method that draws the scale factor (e.g.: 1/200) to the upper right hand
     * corner of the view panel.
     */
    private void drawScaleFactor(Graphics2D g2, Dimension size) {

        int w = size.width;
        //int h = size.height;
        String drawScale = MessageFormat.format(resBundle.getString("topViewScale"), getScale());

        FontMetrics metrics = g2.getFontMetrics();
        int width = metrics.stringWidth(drawScale);
        int height = metrics.getHeight();

        g2.drawString(drawScale, w-20-width, 20+height/2);

    }

    /**
     * Method that returns the root chord length (in meters) if defined, else it
     * returns a default value.
     */
    private float getCrLength() {
        if (crParam.isDefined())
            return (float)crParam.get().doubleValue(SI.METER);
        else
            return DEFAULT_CR;
    }

    /**
     * Method that returns the tip chord length (in meters) if defined, else it
     * returns a default value.
     */
    private float getCtLength() {
        if (ctParam.isDefined())
            return (float)ctParam.get().doubleValue(SI.METER);
        else
            return DEFAULT_CT;
    }

    /**
     * Method that returns the span length (in meters) if defined, else it
     * returns a default value.
     */
    private float getSpanLength() {
        if (bParam.isDefined())
            return (float)bParam.get().doubleValue(SI.METER);
        else
            return DEFAULT_SPAN;
    }

    /**
     * Method that returns the leading edge sweep (in radians) if defined, else
     * it returns a default value.
     */
    private float getSweepLE() {
        if (SweepLEParam.isDefined())
            return (float)SweepLEParam.get().doubleValue(SI.RADIAN);
        else
            return DEFAULT_SWEEPLE;
    }

    /**
     * Method that returns the minimum Y value in the geometry.
     */
    private float getMinY() {
        float yMin = 0;

        float sweep = getSweepLE();
        if (sweep<0)
            yMin = (float)(getSpanLength()/2*Math.tan(sweep));

        return yMin;
    }

    /**
     * Method that returns the maximum Y value in the geometry.
     */
    private float getMaxY() {
        float yMax = getCrLength();

        float sweep = getSweepLE();
        float xtLE = (float)(getSpanLength()/2*Math.tan(sweep));
        float xtTE = xtLE+getCrLength();
        if (xtTE>yMax)
            yMax = xtTE;

        return yMax;
    }

    /**
     * Method that returns the minimum X value in the geometry.
     */
    private float getMinX() {
        return -getMaxX();
    }

    /**
     * Method that returns the maximum X value in the geometry.
     */
    private float getMaxX() {
        return getSpanLength()/2;
    }

    /**
     * Called automatically by the Observable notification system when one of
     * the parameters this object is dependent on has changed.
     *
     * @param o The object being observed (this object)
     * @param arg The object that this one depends on that has changed.
     */
    @Override
    public void update(Observable o, Object arg) {
        this.repaint();
    }
}
