/*
 *  WingGeom  -- A Java Application that calculates wing geometry parameters.
 *  Copyright (C) 1998-2023, Joseph A. Huwaldt
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/gpl.html
 */
import jahuwaldt.io.ExtFilenameFilter;
import jahuwaldt.io.FileUtils;
import jahuwaldt.io.JarUtils;
import jahuwaldt.js.ParamTools.*;
import jahuwaldt.js.param.Parameter;
import jahuwaldt.js.pmethods.winggeom.*;
import jahuwaldt.swing.SplashScreen;
import jahuwaldt.swing.*;
import jahuwaldt.util.BrowserHelp;
import jahuwaldt.util.GeneralFormat;
import java.awt.*;
import java.awt.desktop.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.image.BufferedImage;
import java.io.*;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.MalformedURLException;
import java.net.URL;
import java.text.MessageFormat;
import java.util.List;
import static java.util.Objects.isNull;
import static java.util.Objects.nonNull;
import java.util.*;
import java.util.jar.JarEntry;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.imageio.ImageIO;
import javax.measure.quantity.Area;
import javax.measure.quantity.Dimensionless;
import javax.measure.unit.NonSI;
import javax.measure.unit.SI;
import javax.measure.unit.Unit;
import javax.swing.*;
import javax.swing.border.EmptyBorder;
import javax.swing.event.HyperlinkEvent;
import javax.swing.event.HyperlinkListener;
import javolution.util.FastTable;
import javolution.xml.XMLBinding;
import javolution.xml.XMLFormat;
import javolution.xml.XMLObjectReader;
import javolution.xml.XMLObjectWriter;
import javolution.xml.stream.XMLStreamException;


/**
 * Application to calculate the parameters related to trapezoidal wing geometry.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt Date: January 10, 1998
 * @version December 12, 2023
 */
public class WingGeom extends MDIApplication {
    private static final short  WIN_WIDTH = 640;    //  Application frame width & height.
    private static final short  WIN_HEIGHT = 650;
    
    //  A count of how many windows have been opened.
    private static int _windowCount = 0;
    
    /**
    *  The resource bundle for this application.
    *  The ResourceBundle below contains all of the strings used in this
    *  application.  ResourceBundles are useful for localizing applications.
    *  New localities can be added by adding additional properties files.
    */
    private static final ResourceBundle RESBUNDLE = ResourceBundle.getBundle("appStrings", java.util.Locale.getDefault());
    
    //  The base File for the help files used by this application.
    private static File _helpBase;
    

    //-----------------------------------------------------------------------------------
    /**
     * This is the "main" routine that is called when an application first
     * starts up.
     *
     * @param args Command line arguments. Ignored.
     */
    public static void main(String args[]) {

        // All code that creates or interacts with Swing components must run on the
        // event dispatch thread.
        SwingUtilities.invokeLater(new Runnable() {
            @Override
            @SuppressWarnings("UseSpecificCatch")
            public void run() {
                try {
                    // Set the system look and feel to hide that hideous Java LAF.
                    AppUtilities.setSystemLAF();

                    // MacOS X users like the menu bar to be at the top of the screen.
                    // This does nothing on non-MacOS platforms.
                    System.setProperty("apple.laf.useScreenMenuBar", "true");

                    //  Display splash screen at applications startup.
                    URL resource = WingGeom.class.getResource(RESBUNDLE.getString("splashScreenImg"));
                    new SplashScreen(resource, null, SplashScreen.ONE_SECOND * 4);

                    // Create an instance of our application.
                    new WingGeom();

                } catch (ParamException e) {
                    e.printStackTrace();
                    JOptionPane.showMessageDialog(null, e.getMessage(), RESBUNDLE.getString("paramErrorTitle"),
                            JOptionPane.ERROR_MESSAGE);
                    System.exit(0);

                } catch (MalformedURLException e) {
                    e.printStackTrace();
                    JOptionPane.showMessageDialog(null, e.getMessage(), RESBUNDLE.getString("paramDispErrorTitle"),
                            JOptionPane.ERROR_MESSAGE);
                    System.exit(0);

                } catch (Exception e) {
                    e.printStackTrace();
                    AppUtilities.showException(null, RESBUNDLE.getString("unexpectedTitle"),
                            RESBUNDLE.getString("unexpectedMsg"), e);
                    System.exit(0);
                    
                } catch (Throwable e) {
                    e.printStackTrace();
                    AppUtilities.showException(null, RESBUNDLE.getString("unexpectedTitle"),
                            RESBUNDLE.getString("unexpectedMsg"), e);
                    System.exit(0);
                }
            }
        });

    }

    /**
     * Constructor for our application.
     *
     * @throws java.lang.Exception if there is any problem creating the
     * application.
     */
    @SuppressWarnings("OverridableMethodCallInConstructor")
    public WingGeom() throws Exception {
        super(RESBUNDLE.getString("appName"));

        // Add an open files handler for MacOS.
        setOpenFileHandler(new OpenFilesHandler() {
            @Override
            public void openFiles(OpenFilesEvent e) {
                //  Loop over the files to be opened and open each in turn.
                List<File> files = e.getFiles();
                for (File file : files) {
                    openFile(file);
                }
            }
        });
        
        //  Extract the help files from this application's JAR file into a temporary directory
        //  so that they can be accessed by a web browser.
        _helpBase = extractHelpFiles();

        //  Create a quit listener that deletes our help files which have been saved off.
        addQuitListener(new QuitListener() {
            @Override
            public boolean quit() {
                //  Delete the temporary directory created earlier.
                FileUtils.deleteDirectory(_helpBase);
                return false;
            }
        });

        //  Create the menu bar.
        JMenuBar menuBar = createMenuBar(false);

        // Add the menu bar to this application.
        //  Cause the application to stay open when the last window is closed under MacOS.
        //  When the last window is closed, a menu bar will remain for the application with no
        //  window open.  This is standard MacOS behavior and will do nothing on other platforms.
        setFramelessJMenuBar(menuBar);

        //  Create a main window frame for this application.
        AppWindow window = new AppWindow();
        addWindow(window);
        window.setVisible(true);

    }

    /**
     * Method that extracts the help files from this application's JAR file and
     * saves them in a temporary directory. A reference to the temporary
     * directory is returned.
     */
    private File extractHelpFiles() throws Exception {

        //  Get a reference to the JAR file that this program is running from.
        URL jarURL = this.getClass().getProtectionDomain().getCodeSource().getLocation();

        //  Create a temporary directory.
        File tmpDir = FileUtils.createTempDirectory("WingGeom");

        //  Extract the help directory from inside the JAR into the temporary directory.
        InputStream is = jarURL.openStream();
        JarUtils.extractJar(is, tmpDir, new JarEntry("help"));
        is.close();

        return tmpDir;
    }

    /**
     * Return a reference to the help files as a URL.
     *
     * @param fileName The name of the particular help file requested.
     * @return A URL pointing to the specified help file (which is in a
     * temporary directory as it is extracted from this application's JAR file).
     * @throws java.net.MalformedURLException if there is a problem with the URL.
     */
    public URL getHelpFileURL(String fileName) throws MalformedURLException {
        return new URL(_helpBase.toURI().toURL(), "help/" + fileName);
    }

    /**
     * Handle the user choosing the "New..." from the File menu. Creates a new
     * document window and returns a reference to it.
     * 
     * @param event The event that caused this method to be called.  Ignored.
     * @return The new window instance.
     */
    @Override
    public Frame handleNew(ActionEvent event) {
        AppWindow window = null;
        try {

            // Create an empty instance of this application's window.
            window = new AppWindow();
            addWindow(window);
            window.setVisible(true);

        } catch (ParamException e) {
            e.printStackTrace();
            JOptionPane.showMessageDialog(null, e.getMessage(), RESBUNDLE.getString("paramErrorTitle"),
                    JOptionPane.ERROR_MESSAGE);

        } catch (MalformedURLException e) {
            e.printStackTrace();
            JOptionPane.showMessageDialog(null, e.getMessage(), RESBUNDLE.getString("paramDispErrorTitle"),
                    JOptionPane.ERROR_MESSAGE);

        } catch (Exception e) {
            e.printStackTrace();
            AppUtilities.showException(null, RESBUNDLE.getString("unexpectedTitle"),
                    RESBUNDLE.getString("unexpectedMsg"), e);
        }

        return window;
    }

    /**
     * Handle the user choosing the "Save" option from the File menu. Saves the
     * values of each parameter to an XML file.
     * 
     * @param event The event that caused this method to be called.  Ignored.
     */
    public void handleSave(ActionEvent event) {
        AppWindow topWindow = (AppWindow)getTopWindow();
        if (topWindow != null && MacOSUtilities.isMarkedModified(topWindow)) {
            //  Does the top window have a file associated with it already?
            File oldFile = MacOSUtilities.getProxyIcon(topWindow);
            if (oldFile != null && oldFile.exists() && oldFile.canWrite())
                topWindow.saveFile(oldFile);
            else
                //  Fall back on Save As...
                topWindow.doSaveAs();
        }
    }

    /**
     * Handle the user choosing the "Save As..." from the File menu. Saves the
     * values of each parameter to a newly selected XML file.
     * 
     * @param event The event that caused this method to be called.  Ignored.
     */
    public void handleSaveAs(ActionEvent event) {
        AppWindow topWindow = (AppWindow)getTopWindow();
        if (topWindow != null)
            topWindow.doSaveAs();
    }

    /**
     * Open the specified file in either the top window (if it is empty) or a new window.
     * 
     * @param theFile The file to be opened and read in.
     */
    private void openFile(File theFile) {
        try {
            AppWindow topWindow = (AppWindow) getTopWindow();
            AppWindow window;

            //  Is the top-window empty?
            if (nonNull(topWindow) && topWindow.isEmpty()) {
                //  The top window is empty. Use it.
                window = topWindow;
            } else
                //  Create a new application window and read in the file.
                window = new AppWindow();

            //  Read in the file.
            window.readFile(theFile);

            if (topWindow != window) {
                //  Display the new window.
                addWindow(window);
                window.setVisible(true);
            }

        } catch (ParamException e) {
            e.printStackTrace();
            JOptionPane.showMessageDialog(null, e.getMessage(), RESBUNDLE.getString("paramErrorTitle"),
                    JOptionPane.ERROR_MESSAGE);

        } catch (MalformedURLException e) {
            e.printStackTrace();
            JOptionPane.showMessageDialog(null, e.getMessage(), RESBUNDLE.getString("paramDispErrorTitle"),
                    JOptionPane.ERROR_MESSAGE);

        } catch (IOException e) {
            JOptionPane.showMessageDialog(null, e.toString(), RESBUNDLE.getString("ioErrorTitle"),
                    JOptionPane.ERROR_MESSAGE);

        } catch (Exception e) {
            e.printStackTrace();
            AppUtilities.showException(null, RESBUNDLE.getString("unexpectedTitle"),
                    RESBUNDLE.getString("unexpectedMsg"), e);
        }
         
    }
    
    /**
     * Handle the user choosing "Open..." from the File menu.
     *
     * @param event The event that caused this method to be called. Ignored.
     */
    public void handleOpen(ActionEvent event) {

        AppWindow topWindow = (AppWindow)getTopWindow();
        String dir = null, name = null;
        if (nonNull(topWindow)) {
            //  Is there already a directory to prompt the user with.
            File oldFile = MacOSUtilities.getProxyIcon(topWindow);
            if (nonNull(oldFile))
                dir = oldFile.getParent();
        }
        ExtFilenameFilter fnFilter = new ExtFilenameFilter(RESBUNDLE.getString("extension"));
        String msg = RESBUNDLE.getString("fileDialogLoad");
        File theFile = AppUtilities.selectFile(topWindow, FileDialog.LOAD, msg, dir, name, fnFilter);
        if (nonNull(theFile))
            openFile(theFile);
    }

    /**
     * Handle the user choosing "Save As PNG..." from the File menu. Saves the
     * currently displayed plot as a PNG format image file.
     *
     * @param event The action event that caused this method to be called.
     * Ignored.
     */
    public void handleSaveAsPNG(ActionEvent event) {
        try {
            //	Go off and actually do the Save As PNG....  Catch exceptions here.
            AppWindow topWindow = (AppWindow)getTopWindow();
            topWindow.doSaveAsPNG();

        } catch (IOException e) {
            Logger.getLogger(WingGeom.class.getName()).log(Level.SEVERE, null, e);
            String msg = e.getMessage();
            if (msg == null) {
                Throwable cause = e.getCause();
                if (cause != null)
                    msg = cause.getMessage();
            }
            JOptionPane.showMessageDialog(null, msg, RESBUNDLE.getString("ioErrorTitle"), JOptionPane.ERROR_MESSAGE);

        } catch (Exception e) {
            Logger.getLogger(WingGeom.class.getName()).log(Level.SEVERE, null, e);
            AppUtilities.showException(null, RESBUNDLE.getString("unexpectedTitle"), RESBUNDLE.getString("unexpectedMsg"), e);
        }

    }

    /**
     * Handle the user choosing "Save As SVG..." from the File menu. Saves the
     * currently displayed plot as an SVG format file.
     *
     * @param event The action event that caused this method to be called.
     * Ignored.
     */
    public void handleSaveAsSVG(ActionEvent event) {
        try {
            //	Go off and actually do the Save As SVG....  Catch exceptions here.
            AppWindow topWindow = (AppWindow)getTopWindow();
            topWindow.doSaveAsSVG();

        } catch (IOException e) {
            Logger.getLogger(WingGeom.class.getName()).log(Level.SEVERE, null, e);
            String msg = e.getMessage();
            if (msg == null) {
                Throwable cause = e.getCause();
                if (cause != null)
                    msg = cause.getMessage();
            }
            JOptionPane.showMessageDialog(null, msg, RESBUNDLE.getString("ioErrorTitle"), JOptionPane.ERROR_MESSAGE);

        } catch (Exception e) {
            Logger.getLogger(WingGeom.class.getName()).log(Level.SEVERE, null, e);
            AppUtilities.showException(null, RESBUNDLE.getString("unexpectedTitle"), RESBUNDLE.getString("unexpectedMsg"), e);
        }

    }

    //-----------------------------------------------------------------------------------
    /**
     * Initializes and displays the menus associated with this application.
     */
    private JMenuBar createMenuBar(boolean windowsMenu) throws NoSuchMethodException {

        //  Create a menu bar.
        JMenuBar menuBar = new JMenuBar();

        // Set up the file menu.
        List<String[]> menuStrings = new ArrayList();
        String[] row = new String[3];
        row[0] = RESBUNDLE.getString("newItemText");
        row[1] = RESBUNDLE.getString("newItemKey");
        row[2] = "handleNew";
        menuStrings.add(row);
        row = new String[3];
        row[0] = RESBUNDLE.getString("openItemText");
        row[1] = RESBUNDLE.getString("openItemKey");
        row[2] = "handleOpen";
        menuStrings.add(row);
        menuStrings.add(new String[3]); //  Blank line
        row = new String[3];
        row[0] = RESBUNDLE.getString("closeItemText");
        row[1] = RESBUNDLE.getString("closeItemKey");
        row[2] = "handleClose";
        menuStrings.add(row);
        row = new String[3];
        row[0] = RESBUNDLE.getString("saveItemText");
        row[1] = RESBUNDLE.getString("saveItemKey");
        row[2] = "handleSave";
        menuStrings.add(row);
        row = new String[3];
        row[0] = RESBUNDLE.getString("saveAsItemText");
        row[1] = null;
        row[2] = "handleSaveAs";
        menuStrings.add(row);
        row = new String[3];
        row[0] = RESBUNDLE.getString("saveAsPNGItemText");
        row[1] = null;
        row[2] = "handleSaveAsPNG";
        menuStrings.add(row);
        if (createSVGGraphics2D(10, 10) != null) {
            row = new String[3];
            row[0] = RESBUNDLE.getString("saveAsSVGItemText");
            row[1] = null;
            row[2] = "handleSaveAsSVG";
            menuStrings.add(row);
        }

        JMenu menu = AppUtilities.buildMenu(this, RESBUNDLE.getString("fileMenuText"), menuStrings);
        menuBar.add(menu);
        
        //  Add a Quit menu item (if it isn't automatically present).
        if (!QuitJMenuItem.isAutomaticallyPresent()) {
            menu.addSeparator();
            menu.add(this.getQuitJMenuItem());
        }
        
        //  Add a Quit handler to ask the user if they want to save unsaved changes before quitting.
        setQuitHandler(new QuitHandler() {
            @Override
            public void handleQuitRequestWith​(QuitEvent e, QuitResponse response) {
                handleQuit(e, response);
            }
        });
        
        // Set up the edit menu.
        menuStrings.clear();
        row = new String[3];
        row[0] = RESBUNDLE.getString("undoItemText");
        row[1] = RESBUNDLE.getString("undoItemKey");
        row[2] = null;
        menuStrings.add(row);
        row = new String[3];
        row[0] = RESBUNDLE.getString("redoItemText");
        row[1] = null;
        row[2] = null;
        menuStrings.add(row);
        menuStrings.add(new String[3]); //  Blank line.
        row = new String[3];
        row[0] = RESBUNDLE.getString("cutItemText");
        row[1] = RESBUNDLE.getString("cutItemKey");
        row[2] = null;
        menuStrings.add(row);
        row = new String[3];
        row[0] = RESBUNDLE.getString("copyItemText");
        row[1] = RESBUNDLE.getString("copyItemKey");
        row[2] = null;
        menuStrings.add(row);
        row = new String[3];
        row[0] = RESBUNDLE.getString("pasteItemText");
        row[1] = RESBUNDLE.getString("pasteItemKey");
        row[2] = null;
        menuStrings.add(row);
        menu = AppUtilities.buildMenu(this, RESBUNDLE.getString("editMenuText"), menuStrings);
        menuBar.add(menu);

        //  Create a Window's menu.
        if (windowsMenu) {
            menu = newWindowsMenu(RESBUNDLE.getString("windowsMenuText"));
            menuBar.add(menu);
        }

        //	Register an about menu item handler.
        setAboutHandler(createAboutHandler());
        
        //  Create an about menu item (if not automatically present).
        if (!AboutJMenuItem.isAutomaticallyPresent()) {
            //  Create Help menu.
            menu = new JMenu(RESBUNDLE.getString("helpMenuText"));
            menuBar.add(menu);
            
            //  Add the "About" item to the Help menu.
            menu.add(this.getAboutJMenuItem());
        }
        
        return menuBar;
    }

    /**
     * Create an about handler for use in this application.
     * 
     * @return the about menu item for this application.
     */
    public AboutHandler createAboutHandler() {

        AboutHandler about = new AboutHandler() {
            @Override
            public void handleAbout​(AboutEvent e) {

                //  Load in the application's icon image.
                Icon appIcon = null;

                try {
                    URL imgURL = ClassLoader.getSystemResource(RESBUNDLE.getString("applicationIconURL"));
                    appIcon = new ImageIcon(imgURL);
                } catch (Exception ex) {
                    ex.printStackTrace();
                }

                if (isNull(appIcon))
                    //  Use a generic icon since this app can't read it's custom icon.
                    appIcon = UIManager.getIcon("OptionPane.informationIcon");

                //  Read in the about box text information.
                String credits = readAboutText(RESBUNDLE.getString("aboutTextURLStr"));

                String aName = RESBUNDLE.getString("appName");
                String aVersion = RESBUNDLE.getString("appVersion");
                String aModDate = RESBUNDLE.getString("appModDate");
                String copyright = RESBUNDLE.getString("copyright");
                StandardMacAboutFrame f = new StandardMacAboutFrame(aName, aVersion + " - " + aModDate);
                f.setApplicationIcon(appIcon);
                f.setCopyright(copyright);
                f.setCredits(credits, "text/html");
                f.setHyperlinkListener(new HyperlinkListener() {
                    @Override
                    public void hyperlinkUpdate(HyperlinkEvent e) {
                        if (e.getEventType() == HyperlinkEvent.EventType.ACTIVATED) {
                            //  Display the selected URL in the default browser for this platform.
                            try {
                                MDIApplication.browse(e.getURL().toURI());
                            } catch (Exception err) {
                                err.printStackTrace();
                            }
                        }
                    }
                });
                f.setVisible(true);
            }
        };

        return about;
    }

    /**
     * Method that handles reading in the contents of the text region of the
     * about box from a text file.
     */
    private String readAboutText(String urlStr) {
        String text;
        try {

            URL url = ClassLoader.getSystemResource(urlStr);

            InputStreamReader reader = new InputStreamReader(url.openStream(), "UTF-8");
            StringWriter writer = new StringWriter();
            int character = reader.read();
            while (character != -1) {
                writer.write(character);
                character = reader.read();
            }
            text = writer.toString();

        } catch (IOException e) {
            e.printStackTrace();

            //  Create a standard set of credits if we didn't find the text file.
            text = "<html><body><b>Author:</b><br>Joseph A. Huwaldt<br>";
            text += "<a href=\"mailto:jhuwaldt@mac.com\">jhuwaldt@mac.com</a><br>";
            text += "<P ALIGN=CENTER><BR>" + getResourceBundle().getString("appName") + " comes with ABSOLUTELY NO WARRANTY;";
            text += "</body></html>";
        }

        return text;
    }

    //-----------------------------------------------------------------------------------
    /**
     * The main window frame for this application.
     */
    private class AppWindow extends JFrame implements Observer {

        //  A database to hold our parameters.
        private ParamDatabase _pdb = new ParamDatabase();

        //  The object that generates the geometry view for this window.
        private TopViewPanel _geomView;

        //  Flag indicating that this window has no contents if true.
        private boolean isEmpty = true;

        //  Is the application in the process of quitting?
        private boolean isQuitting = false;

        /**
         * Construct a new Untitled AppWindow instance.
         */
        @SuppressWarnings({"LeakingThisInConstructor", "OverridableMethodCallInConstructor"})
        public AppWindow() throws Exception {
            super(RESBUNDLE.getString("untitled") + (++_windowCount > 1 ? " " + String.valueOf(_windowCount) : ""));

            //  Have the window do nothing automatically when it closes.
            //  We handle closing with a WindowListener.
            this.setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);

            //  Create a menu bar for this window.
            JMenuBar menuBar = createMenuBar(true);
            setJMenuBar(menuBar);

            //  Initialize the parameters used in this program.
            initParams();

            //  Display input boxes for all the parameters in this window.
            List<String> pLabels = new ArrayList();
            pLabels.add("bw");
            pLabels.add("ARw");
            pLabels.add("Sw");
            pLabels.add("Crw");
            pLabels.add("Ctw");
            pLabels.add("taperw");
            pLabels.add("SweepLEw");
            pLabels.add("Sweep1/4cw");
            pLabels.add("SweepTEw");
            pLabels.add("MACw");
            pLabels.add("EtaMACw");
            pLabels.add("yMACw");
            pLabels.add("xMACw");
            pLabels.add("xMAC4w");

            //  Layout the content of the frame.
            Container cp = this.getContentPane();
            cp.setLayout(new BorderLayout());

            // Create an info region at the top.
            cp.add(layoutInfoRgn(RESBUNDLE.getString("appName"),
                    MessageFormat.format(RESBUNDLE.getString("infoRgnLabel1Msg"), RESBUNDLE.getString("appVersion")),
                    MessageFormat.format(RESBUNDLE.getString("infoRgnLabel2Msg"), RESBUNDLE.getString("appAuthor")),
                    RESBUNDLE.getString("appModDate")), BorderLayout.NORTH);

            //  Add the input parameters to be displayed to the window.
            cp.add(new ParamDBInputPanel(_pdb, pLabels, 3, RESBUNDLE, null), BorderLayout.CENTER);

            //  Create and link the geometry generator to the appropriate parameters.
            _geomView = new TopViewPanel(RESBUNDLE, (AutoParam)_pdb.get("Crw"), (AutoParam)_pdb.get("Ctw"),
                    (AutoParam)_pdb.get("bw"), (AutoParam)_pdb.get("SweepLEw"));
            _geomView.setPreferredSize(new Dimension(640, 250));

            //  Add the geometry pane to the lower part of the window.
            cp.add(_geomView, BorderLayout.SOUTH);

            //  Size the window and position it.
            AppUtilities.positionWindow(this, WIN_WIDTH, WIN_HEIGHT);

            //  Add a window close listener that asks the user to save any unsaved data.
            addWindowListener(new WindowAdapter() {
                @Override
                public void windowClosing(WindowEvent e) {
                    if (!isQuitting && MacOSUtilities.isMarkedModified(AppWindow.this)) {
                        //  Ask the user what they want to do.
                        int option = showSaveChangesDialog();
                        if (option == JOptionPane.YES_OPTION) {
                            //  Save the file.
                            doSaveAs();

                        } else if (option == JOptionPane.CANCEL_OPTION)
                            return;
                    }
                    //  Go ahead and close the window.
                    AppWindow.this.dispose();
                }
            });

            //  Create a quit listener that asks the user to save any unsaved data.
            addQuitListener(new QuitListener() {
                @Override
                public boolean quit() {
                    isQuitting = true;
                    //  Are there any unsaved changes?
                    if (AppWindow.this.isVisible() &&  MacOSUtilities.isMarkedModified(AppWindow.this)) {
                        //  Ask the user what they want to do.
                        int option = showSaveChangesDialog();
                        if (option == JOptionPane.YES_OPTION) {
                            //  Save the file.
                            doSaveAs();

                        } else if (option == JOptionPane.CANCEL_OPTION) {
                            isQuitting = false;
                            return true;    //  Request that the quit be canceled.
                        }
                    }
                    //  Go ahead and quit the application.
                    return false;
                }
            });

        }

        /**
         * Present a dialog asking the user if they want to save any changes.
         *
         * @return The JOptionPane.YES_NO_CANCEL_OPTION option.
         */
        private int showSaveChangesDialog() {
            AppWindow.this.setVisible(true);
            Object[] options = {RESBUNDLE.getString("saveChangesSave"),
                RESBUNDLE.getString("dontSaveChanges"),
                RESBUNDLE.getString("cancel")};
            int option = JOptionPane.showOptionDialog(AppWindow.this,
                    RESBUNDLE.getString("saveChangesMsg"),
                    RESBUNDLE.getString("selectOptionTitle"),
                    JOptionPane.YES_NO_CANCEL_OPTION, JOptionPane.QUESTION_MESSAGE,
                    null, options, options[0]);
            return option;
        }

        /**
         * Sets the title for this frame to the specified string. Also notifies
         * the main application class so that the "Windows" menu gets updated
         * too.
         *
         * @param title The title to be displayed in the frame's border. A null
         * value is treated as an empty string, "".
         */
        @Override
        public void setTitle(String title) {
            super.setTitle(title);
            MDIApplication.windowTitleChanged(this);
        }

        /**
         * Creates and lays out a panel that contains the header information and
         * "clear all" button.
         *
         * @param label1 The primary message to display in the info region.
         * @param label2 The secondary message to display in the info region.
         * @return A panel containing the header information and clear all
         * button.
         */
        private JPanel layoutInfoRgn(String label1, String label2, String label3, String label4) {

            JPanel infoRgnPanel = new JPanel(new BorderLayout());

            // Place information labels.
            Box infoPanel1 = new Box(BoxLayout.Y_AXIS);
            JLabel infoLabel = new JLabel(label1);
            infoPanel1.add(infoLabel);
            infoPanel1.add(Box.createVerticalStrut(5));
            infoLabel = new JLabel(label2);
            infoPanel1.add(infoLabel);

            Box infoPanel2 = new Box(BoxLayout.Y_AXIS);
            infoLabel = new JLabel(label3);
            infoPanel2.add(infoLabel);
            infoPanel2.add(Box.createVerticalStrut(5));
            infoLabel = new JLabel(label4);
            infoPanel2.add(infoLabel);

            Box infoPanel3 = new Box(BoxLayout.X_AXIS);
            infoPanel3.add(infoPanel1);
            infoPanel3.add(Box.createHorizontalStrut(40));
            infoPanel3.add(infoPanel2);

            infoRgnPanel.add(infoPanel3, BorderLayout.WEST);

            //  Place Clear button.
            JButton clearAllButton = new JButton(RESBUNDLE.getString("clearAllBtnStr"));
            clearAllButton.addActionListener(new ActionListener() {
                @Override
                public void actionPerformed(ActionEvent evt) {
                    AutoParam.clearAllParams(_pdb);
                }
            });
            clearAllButton.setToolTipText(RESBUNDLE.getString("clearAllToolTip"));
            infoRgnPanel.add(clearAllButton, BorderLayout.EAST);

            infoRgnPanel.setBorder(new EmptyBorder(10, 20, 10, 20));

            return infoRgnPanel;
        }

        /**
         * Initialize all the parameter objects used by this program.
         */
        private void initParams() throws Exception {
            GeneralFormat nf = new GeneralFormat();

            Unit[] areaUnits = {
                NonSI.SQUARE_FOOT,
                NonSI.SQUARE_INCH,
                SI.SQUARE_METRE,
                SI.CENTIMETER.times(SI.CENTIMETER).asType(Area.class)};
            Unit[] lengthUnits = {
                NonSI.FOOT,
                NonSI.INCH,
                SI.METER,
                SI.CENTIMETER};
            Unit[] angularUnits = {NonSI.DEGREE_ANGLE, SI.RADIAN};

            //  Get reference to the Help button image.
            URL helpIconURL = getResource("helpIconImg");

            //  Create an object to contain the inputs we are using.
            AutoParamInputs inputs = new AutoParamInputs();
            inputs.theDatabase = _pdb;

            //  bw parameter
            //  Fill in the inputs data structure.
            inputs.label = "bw";
            inputs.desc = RESBUNDLE.getString("bwToolTip");
            inputs.symbolURL = getResource("bwLabelImg");
            inputs.theUnits = NonSI.FOOT;
            inputs.availUnits = lengthUnits;
            inputs.verifierMethod = new bVerifier();
            //  Create the parameter object (automatically stored in _pdb through inputs.theDatabase).
            AutoParam aParam = new AutoParam(inputs);
            //  Add a number formatter for this parameter.
            aParam.setNumberFormat(nf);
            //  Add help information for this parameter.
            aParam.setHelp(helpIconURL, getHelp("bw.html"));
            //  Add calculation methods.
            aParam.addMethod(new bfn_S_AR("Sw", "ARw"));
            aParam.addMethod(new bfn_S_Cr_Ct("Sw", "Crw", "Ctw"));
            aParam.addMethod(new bfn_AR_Cr_Ct("ARw", "Crw", "Ctw"));
            aParam.addMethod(new bfn_yMAC_etaMAC("yMACw", "EtaMACw"));
            aParam.addObserver(this);   //  Add this window as an observer.

            //  ARw parameter
            inputs.label = "ARw";
            inputs.desc = RESBUNDLE.getString("ARwToolTip");
            inputs.symbolURL = getResource("ARLabelImg");
            inputs.theUnits = Dimensionless.UNIT;
            inputs.availUnits = null;
            inputs.verifierMethod = new ARVerifier();
            aParam = new AutoParam(inputs);
            aParam.setNumberFormat(nf);
            aParam.setHelp(helpIconURL, getHelp("ARw.html"));
            aParam.addMethod(new ARfn_b_S("bw", "Sw"));
            aParam.addMethod(new ARfn_b_Cr_Ct("bw", "Crw", "Ctw"));
            aParam.addMethod(new ARfn_sweepLE_sweep4_taper("SweepLEw", "Sweep1/4cw", "taperw"));
            aParam.addMethod(new ARfn_sweepLE_sweepTE_taper("SweepLEw", "SweepTEw", "taperw"));
            aParam.addObserver(this);   //  Add this window as an observer.

            //  Sw parameter
            inputs.label = "Sw";
            inputs.desc = RESBUNDLE.getString("SwToolTip");
            inputs.symbolURL = getResource("SwLabelImg");
            inputs.theUnits = NonSI.SQUARE_FOOT;
            inputs.availUnits = areaUnits;
            inputs.verifierMethod = new SVerifier();
            aParam = new AutoParam(inputs);
            aParam.setNumberFormat(nf);
            aParam.setHelp(helpIconURL, getHelp("Sw.html"));
            aParam.addMethod(new Sfn_b_AR("bw", "ARw"));
            aParam.addMethod(new Sfn_b_Cr_Ct("bw", "Crw", "Ctw"));
            aParam.addObserver(this);   //  Add this window as an observer.

            //  Crw parameter
            inputs.label = "Crw";
            inputs.desc = RESBUNDLE.getString("CrwToolTip");
            inputs.symbolURL = getResource("CrwLabelImg");
            inputs.theUnits = NonSI.FOOT;
            inputs.availUnits = lengthUnits;
            inputs.verifierMethod = new ChordVerifier();
            aParam = new AutoParam(inputs);
            aParam.setNumberFormat(nf);
            aParam.setHelp(helpIconURL, getHelp("Crw.html"));
            aParam.addMethod(new Crfn_S_b_Ct("Sw", "bw", "Ctw"));
            aParam.addMethod(new Crfn_taper_Ct("taperw", "Ctw"));
            aParam.addMethod(new Crfn_S_b_taper("Sw", "bw", "taperw"));
            aParam.addMethod(new Crfn_Cbar_taper("MACw", "taperw"));
            aParam.addObserver(this);   //  Add this window as an observer.

            //  Ctw parameter
            inputs.label = "Ctw";
            inputs.desc = RESBUNDLE.getString("CtwToolTip");
            inputs.symbolURL = getResource("CtwLabelImg");
            inputs.theUnits = NonSI.FOOT;
            inputs.availUnits = lengthUnits;
            inputs.verifierMethod = new ChordVerifier();
            aParam = new AutoParam(inputs);
            aParam.setNumberFormat(nf);
            aParam.setHelp(helpIconURL, getHelp("Ctw.html"));
            aParam.addMethod(new Ctfn_S_b_Cr("Sw", "bw", "Crw"));
            aParam.addMethod(new Ctfn_taper_Cr("taperw", "Crw"));
            aParam.addObserver(this);   //  Add this window as an observer.

            //  Taper ratio parameter
            inputs.label = "taperw";
            inputs.desc = RESBUNDLE.getString("taperwToolTip");
            inputs.symbolURL = getResource("taperwLabelImg");
            inputs.theUnits = Dimensionless.UNIT;
            inputs.availUnits = null;
            inputs.verifierMethod = new taperVerifier();
            aParam = new AutoParam(inputs);
            aParam.setNumberFormat(nf);
            aParam.setHelp(helpIconURL, getHelp("taperw.html"));
            aParam.addMethod(new taperfn_Cr_Ct("Crw", "Ctw"));
            aParam.addMethod(new taperfn_Cr_Cbar("Crw", "MACw"));
            aParam.addMethod(new taperfn_etaMAC("EtaMACw"));
            aParam.addMethod(new taperfn_swpLE_swp4_AR("SweepLEw", "Sweep1/4cw", "ARw"));
            aParam.addMethod(new taperfn_swpLE_swpTE_AR("SweepLEw", "SweepTEw", "ARw"));
            aParam.addObserver(this);   //  Add this window as an observer.

            //  MACw parameter
            inputs.label = "MACw";
            inputs.desc = RESBUNDLE.getString("MACwToolTip");
            inputs.symbolURL = getResource("MACwLabelImg");
            inputs.theUnits = NonSI.FOOT;
            inputs.availUnits = lengthUnits;
            inputs.verifierMethod = new ChordVerifier();
            aParam = new AutoParam(inputs);
            aParam.setNumberFormat(nf);
            aParam.setHelp(helpIconURL, getHelp("Cbarw.html"));
            aParam.addMethod(new Cbarfn_Cr_taper("Crw", "taperw"));
            aParam.addObserver(this);   //  Add this window as an observer.

            //  EtaMACw parameter
            inputs.label = "EtaMACw";
            inputs.desc = RESBUNDLE.getString("EtaMACwToolTip");
            inputs.symbolURL = getResource("etaMACwLabelImg");
            inputs.theUnits = Dimensionless.UNIT;
            inputs.availUnits = null;
            inputs.verifierMethod = new EtaMACVerifier();
            aParam = new AutoParam(inputs);
            aParam.setNumberFormat(nf);
            aParam.setHelp(helpIconURL, getHelp("etaMACw.html"));
            aParam.addMethod(new etaMACfn_taper("taperw"));
            aParam.addMethod(new etaMACfn_yMAC_b("yMACw", "bw"));
            aParam.addObserver(this);   //  Add this window as an observer.

            //  yMACw parameter
            inputs.label = "yMACw";
            inputs.desc = RESBUNDLE.getString("yMACwToolTip");
            inputs.symbolURL = getResource("yMACwLabelImg");
            inputs.theUnits = NonSI.FOOT;
            inputs.availUnits = lengthUnits;
            inputs.verifierMethod = new yMACVerifier();
            aParam = new AutoParam(inputs);
            aParam.setNumberFormat(nf);
            aParam.setHelp(helpIconURL, getHelp("yMACw.html"));
            aParam.addMethod(new yMACfn_etaMAC_b("EtaMACw", "bw"));
            aParam.addMethod(new yMACfn_xMAC_sweepLE("xMACw", "SweepLEw"));
            aParam.addObserver(this);   //  Add this window as an observer.

            //  xMACw parameter
            inputs.label = "xMACw";
            inputs.desc = RESBUNDLE.getString("xMACwToolTip");
            inputs.symbolURL = getResource("xMACwLabelImg");
            inputs.theUnits = NonSI.FOOT;
            inputs.availUnits = lengthUnits;
            inputs.verifierMethod = null;
            aParam = new AutoParam(inputs);
            aParam.setNumberFormat(nf);
            aParam.setHelp(helpIconURL, getHelp("xMACw.html"));
            aParam.addMethod(new xMACfn_yMAC_sweepLE("yMACw", "SweepLEw"));
            aParam.addMethod(new xMACfn_xMAC4_Cbar("xMAC4w", "MACw"));
            aParam.addObserver(this);   //  Add this window as an observer.

            //  xMAC4w parameter
            inputs.label = "xMAC4w";
            inputs.desc = RESBUNDLE.getString("xMAC4wToolTip");
            inputs.symbolURL = getResource("xMAC4wLabelImg");
            inputs.theUnits = NonSI.FOOT;
            inputs.availUnits = lengthUnits;
            inputs.verifierMethod = null;
            aParam = new AutoParam(inputs);
            aParam.setNumberFormat(nf);
            aParam.setHelp(helpIconURL, getHelp("xMAC4w.html"));
            aParam.addMethod(new xMAC4fn_xMAC_Cbar("xMACw", "MACw"));
            aParam.addObserver(this);   //  Add this window as an observer.

            //  SweepLEw parameter
            inputs.label = "SweepLEw";
            inputs.desc = RESBUNDLE.getString("SweepLEwToolTip");
            inputs.symbolURL = getResource("sweepLEwLabelImg");
            inputs.theUnits = NonSI.DEGREE_ANGLE;
            inputs.availUnits = angularUnits;
            inputs.verifierMethod = null;
            aParam = new AutoParam(inputs);
            aParam.setNumberFormat(nf);
            aParam.setHelp(helpIconURL, getHelp("sweepLEw.html"));
            aParam.addMethod(new sweepLEfn_sweep4_AR_taper("Sweep1/4cw", "ARw", "taperw"));
            aParam.addMethod(new sweepLEfn_sweepTE_AR_taper("SweepTEw", "ARw", "taperw"));
            aParam.addMethod(new sweepLEfn_xMAC_yMAC("xMACw", "yMACw"));
            aParam.addObserver(this);   //  Add this window as an observer.

            //  Sweep1/4cw parameter
            inputs.label = "Sweep1/4cw";
            inputs.desc = RESBUNDLE.getString("Sweep4cwToolTip");
            inputs.symbolURL = getResource("sweep4wLabelImg");
            inputs.theUnits = NonSI.DEGREE_ANGLE;
            inputs.availUnits = angularUnits;
            inputs.verifierMethod = null;
            aParam = new AutoParam(inputs);
            aParam.setNumberFormat(nf);
            aParam.setHelp(helpIconURL, getHelp("sweep4w.html"));
            aParam.addMethod(new sweep4fn_sweepLE_AR_taper("SweepLEw", "ARw", "taperw"));
            aParam.addObserver(this);   //  Add this window as an observer.

            //  SweepTEw parameter
            inputs.label = "SweepTEw";
            inputs.desc = RESBUNDLE.getString("SweepTEwToolTip");
            inputs.symbolURL = getResource("sweeTEwLabelImg");
            inputs.theUnits = NonSI.DEGREE_ANGLE;
            inputs.availUnits = angularUnits;
            inputs.verifierMethod = null;
            aParam = new AutoParam(inputs);
            aParam.setNumberFormat(nf);
            aParam.setHelp(helpIconURL, getHelp("sweepTEw.html"));
            aParam.addMethod(new sweepTEfn_sweepLE_AR_taper("SweepLEw", "ARw", "taperw"));
            aParam.addObserver(this);   //  Add this window as an observer.

            //  Have each parameter link up with the parameters that it depends on.
            AutoParam.linkAllParams(_pdb);
        }

        /**
         * Method that uses the Java class loader to return the URL to a
         * resource with the specified resource bundle item name.
         */
        private URL getResource(String resName) {
            return this.getClass().getResource(RESBUNDLE.getString(resName));
        }

        /**
         * Method that returns a help method which loads the specified web page
         * from the "help" sub-directory in the application's JAR file.
         */
        private BrowserHelp getHelp(String fileName) throws IOException {
            URL helpURL = getHelpFileURL(fileName);
            BrowserHelp help = new BrowserHelp(helpURL);
            return help;
        }

        /**
         * Called automatically by the Observable notification system when one
         * of the parameters this window is dependent on has changed.
         *
         * @param o The object being observed
         * @param arg The object that this window depends on that has changed.
         */
        @Override
        public void update(Observable o, Object arg) {
            //  Mark the window as having been modified.
            MacOSUtilities.setModifiedMark(this, true);
            isEmpty = false;
        }

        /**
         * Returns true if this window is blank or empty.
         */
        public boolean isEmpty() {
            return isEmpty;
        }

        /**
         * Asks the user to select an output file and writes the parameters to
         * it.
         */
        void doSaveAs() {
            // Bring up a file chooser.
            String extension = RESBUNDLE.getString("extension");
            ExtFilenameFilter fnFilter = new ExtFilenameFilter(extension);
            String msg = RESBUNDLE.getString("fileSaveDialogMsg");
            String name = RESBUNDLE.getString("untitled");
            String dir = null;
            File theFile = MacOSUtilities.getProxyIcon(this);
            if (theFile != null) {
                name = theFile.getName();
                dir = theFile.getParent();
            }
            theFile = AppUtilities.selectFile4Save(this, msg, dir, name,
                    fnFilter, extension, RESBUNDLE.getString("fileExistsMsgFmt"), RESBUNDLE.getString("warningTitle"));

            if (theFile != null) {
                // Write out the values of all the parameters to an XML file.
                saveFile(theFile);
            }
        }

        /**
         * Method that writes out all the parameters in the parameter database
         * to the specified XML file.
         */
        public void saveFile(File theFile) {

            try {
                //  Create a list of comments (for human reader consumption).
                FastTable<String> comments = FastTable.newInstance();
                String msg = MessageFormat.format(RESBUNDLE.getString("fileCommentMsg1"),
                        RESBUNDLE.getString("appName"), RESBUNDLE.getString("appVersion"));
                comments.add(msg);
                Date date = new Date();
                comments.add(MessageFormat.format(RESBUNDLE.getString("fileCommentMsg2"), date,date));

                //  Sort the parameter database into input and output parameters.
                ParamDatabase inputs = new ParamDatabase();
                ParamDatabase outputs = new ParamDatabase();
                for (Param param : _pdb) {
                    AutoParam aParam = (AutoParam)param;
                    if (aParam.isManuallyInput())
                        inputs.put(aParam.getLabel(), aParam);
                    else
                        outputs.put(aParam.getLabel(), aParam);
                }

                //  Create an XML writer accessing the file.
                XMLObjectWriter writer = XMLObjectWriter.newInstance(new FileOutputStream(theFile));
                writer.setBinding(new WingGeomXMLBinding());
                writer.setIndentation("\t");

                //  Write out the comments.
                writer.write(comments, "Comments", FastTable.class);

                //  Write out the input parameters.
                writer.write(inputs, "Inputs", ParamDatabase.class);

                //  Write out the output parameters.
                writer.write(outputs, "Outputs", ParamDatabase.class);

                //  Close the output stream.
                writer.close();

                //  Change the title of the window to the saved file name.
                this.setTitle(theFile.getName());

                //  Change the window decorations.
                MacOSUtilities.setModifiedMark(this, false);
                MacOSUtilities.setProxyIcon(this, theFile);
                isEmpty = false;

            } catch (IOException e) {
                JOptionPane.showMessageDialog(this, e.toString(), RESBUNDLE.getString("ioErrorTitle"),
                        JOptionPane.ERROR_MESSAGE);

            } catch (XMLStreamException e) {
                e.printStackTrace();
                AppUtilities.showException(null, RESBUNDLE.getString("unexpectedTitle"),
                    RESBUNDLE.getString("unexpectedMsg"), e);
             }
        }

        /**
         * Read in a wgeom file and store the input parameters in this window's
         * parameter database.
         *
         * @param theFile A reference to the wgeom file being read in.
         */
        void readFile(File theFile) throws IOException, XMLStreamException {

            //  Create an XML reader accessing the file.
            XMLObjectReader reader = XMLObjectReader.newInstance(new FileInputStream(theFile));
            reader.setBinding(new WingGeomXMLBinding());

            //  Read in the comment section (but ignore it).
            reader.read("Comments", FastTable.class);

            //  Read in the input parameters (which get stored in _pdb by the WingGeomXMLBinding class).
            reader.read("Inputs", ParamDatabase.class);

            //  Close the input stream.
            reader.close();

            //  Change the title of the window to the save file name.
            this.setTitle(theFile.getName());

            //  Change the window decorations.
            MacOSUtilities.setModifiedMark(this, false);
            MacOSUtilities.setProxyIcon(this, theFile);
        }

        /**
         * Handle the user choosing "Save As PNG..." from the File menu. This asks the
         * user for input and then saves the file. This method throws all exceptions.
         *
         * @return true if the user cancels, false if the file was saved.
         */
        private boolean doSaveAsPNG() throws IOException {

            //	Build the filename to prompt the user with.
            String name = RESBUNDLE.getString("untitled");
            String dir = null;
            File theFile = MacOSUtilities.getProxyIcon(this);
            if (theFile != null) {
                name = theFile.getName();
                dir = theFile.getParent();
            }

            String extension = "png";
            ExtFilenameFilter fnFilter = new ExtFilenameFilter();
            fnFilter.addExtension(extension);

            //	Ask the user to select a file for saving.
            theFile = AppUtilities.selectFile4Save(this,
                    MessageFormat.format(RESBUNDLE.getString("fileSaveDialogMsg"), extension.toUpperCase()),
                    dir, name, fnFilter, extension,
                    RESBUNDLE.getString("fileExistsMsgFmt"), RESBUNDLE.getString("warningTitle"));

            if (theFile != null) {
                if (theFile.exists() && !theFile.canWrite())
                    throw new IOException(RESBUNDLE.getString("canNotWrite2File"));

                //	Write out the PNG file.
                Container c = _geomView;
                BufferedImage im = new BufferedImage(c.getWidth(), c.getHeight(), BufferedImage.TYPE_INT_ARGB);
                c.paint(im.getGraphics());
                ImageIO.write(im, "PNG", theFile);

                return false;
            }

            return true;
        }

        /**
         * Handle the user choosing "Save As SVG..." from the File menu. This asks the
         * user for input and then saves the file. This method throws all exceptions.
         *
         * @return true if the user cancels, false if the file was saved.
         */
        private boolean doSaveAsSVG() throws IOException {

            String name = RESBUNDLE.getString("untitled");
            String dir = null;
            File theFile = MacOSUtilities.getProxyIcon(this);
            if (theFile != null) {
                name = theFile.getName();
                dir = theFile.getParent();
            }

            String extension = "svg";
            ExtFilenameFilter fnFilter = new ExtFilenameFilter();
            fnFilter.addExtension(extension);

            //	Ask the user to select a file for saving.
            theFile = AppUtilities.selectFile4Save(this,
                    MessageFormat.format(RESBUNDLE.getString("fileSaveDialogMsg"), extension.toUpperCase()),
                    dir, name, fnFilter, extension,
                    RESBUNDLE.getString("fileExistsMsgFmt"), RESBUNDLE.getString("warningTitle"));

            if (theFile != null) {
                if (theFile.exists() && !theFile.canWrite())
                    throw new IOException(RESBUNDLE.getString("canNotWrite2File"));

                //	Write out the SVG file.
                try (FileOutputStream outputStream = new FileOutputStream(theFile)) {
                    savePanelAsSVG(_geomView, outputStream);
                }

                return false;
            }

            return true;
        }

        /**
         * Write .svg file based on provided TopViewPanel.
         *
         * @param panel       The top view panel to be written out as SVG.
         * @param outputStream An OutputStream to a file that will hold the SVG XML.
         * @throws java.io.IOException if there is any problem writing out the SVG file.
         */
        public void savePanelAsSVG(TopViewPanel panel, OutputStream outputStream) throws IOException {
            int aWidth = panel.getWidth();
            int aHeight = panel.getHeight();

            // Use reflection to get the SVG string
            String svg = generateSVG(panel, aWidth, aHeight);
            if (svg == null)
                throw new IOException("Problem with creating the SVG plot.");

            PrintWriter output = new PrintWriter(new OutputStreamWriter(outputStream, "UTF-8"));
            output.println("<!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 1.1//EN\" \"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd\">");
            output.println(svg);
            output.flush();

        }

        /**
         * A custom XMLBinding that processes ParamDatabase and Param objects.
         */
        private class WingGeomXMLBinding extends XMLBinding {

            /**
             * A custom XML format for Param objects that saves only the label
             * and the underlying Parameter object. Everything else in a Param
             * instance in this application is invariant (can't be changed by
             * the user). New instances are returned from the already existing
             * parameter database associated with this window and when data is
             * read in from XML the already existing parameters are modified
             * with the values from the file.
             */
            private final XMLFormat<Param> paramXML = new XMLFormat<Param>(null) {
                @Override
                public Param newInstance(Class<Param> cls, XMLFormat.InputElement xml) throws XMLStreamException {
                    String label = xml.getAttribute("label", "");

                    //  Return an already existing Param object with the specified label.
                    return _pdb.get(label);
                }

                @Override
                public void read(XMLFormat.InputElement xml, Param obj) throws XMLStreamException {
                    //  Read in the Parameter instance.
                    Parameter p = (Parameter)xml.getNext();

                    //  Modify the input Param object with the new Parameter value.
                    AutoParam param = (AutoParam)obj;
                    param.makeManuallyInput();
                    param.set(p);
                    param.updateChildren();
                }

                @Override
                public void write(Param obj, XMLFormat.OutputElement xml) throws XMLStreamException {
                    //  Write out the label as an attribute.
                    xml.setAttribute("label", obj.getLabel());

                    //  Write out the underlying Parameter object as the data.
                    xml.add(obj.get());
                }
            };

            /**
             * Set up an XML Binding they way we want for this application.
             */
            public WingGeomXMLBinding() {
                setAlias(String.class, "String");
                setAlias(Parameter.class, "Parameter");
                setAlias(AutoParam.class, "Param");
            }

            /**
             * Override the getFormat() method in order to provide a custom
             * formatter for any Param type objects.
             */
            @Override
            public <T> XMLFormat<T> getFormat(Class<T> cls) {
                if (Param.class.isAssignableFrom(cls))
                    return (XMLFormat<T>)paramXML;
                return super.getFormat(cls);
            }
        }   //  end class WingGeomXMLBinding

    }   //  end class AppWindow

    /**
     * Generates a string containing a rendering of the panel in SVG format. This feature
     * is only supported if the JFreeSVG library is included on the classpath.
     *
     * @return A string containing an SVG element for the specified panel, or
     * <code>null</code> if there is a problem with the method invocation by reflection.
     */
    private static String generateSVG(TopViewPanel panel, int width, int height) {
        Graphics2D g2 = createSVGGraphics2D(width, height);
        if (g2 == null) {
            throw new IllegalStateException("The JFreeSVG library could not be found.");
        }
        String svg = null;
        panel.paintComponent(g2);
        try {
            Method m = g2.getClass().getMethod("getSVGElement");
            svg = (String) m.invoke(g2);
        } catch (NoSuchMethodException | SecurityException | IllegalAccessException | IllegalArgumentException | InvocationTargetException ex) {
            // null will be returned
        }

        return svg;
    }

    private static Graphics2D createSVGGraphics2D(int w, int h) {
        try {
            Class svgGraphics2d = Class.forName("org.jfree.graphics2d.svg.SVGGraphics2D");
            Constructor ctor = svgGraphics2d.getConstructor(int.class, int.class);
            return (Graphics2D) ctor.newInstance(w, h);
        } catch (ClassNotFoundException | NoSuchMethodException | SecurityException | InstantiationException
                | IllegalAccessException | IllegalArgumentException | InvocationTargetException ex) {
            return null;
        }
    }

}
