/*
 *   ParamDatabase - A database or collection of Param objects.
 *
 *   Copyright (C) 1998-2016, by Joseph A. Huwaldt.
 *   All rights reserved.
 *   
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/lgpl.html
 */
package jahuwaldt.js.ParamTools;

import java.io.IOException;
import java.io.PrintStream;
import java.io.Serializable;
import java.util.*;
import java.util.Map.Entry;
import static java.util.Objects.requireNonNull;
import javolution.util.FastMap;
import javolution.xml.XMLFormat;
import javolution.xml.XMLSerializable;
import javolution.xml.stream.XMLStreamException;

/**
 * <p>
 * A database of {@link Param} objects for use in my engineering programs.
 * </p>
 *
 * <p>
 * This database or collection can be used to store parameter objects used by a program.
 * The database is keyed off the parameter label. Therefore, all parameter labels must be
 * unique or new ones will overwrite old ones of the same name. Also, once you've placed a
 * parameter in a ParamDatabase, you must not change it's label.
 * </p>
 *
 * <p> Modified by: Joseph A. Huwaldt </p>
 *
 * @author Joseph A. Huwaldt, Date: January 10, 1998
 * @version March 28, 2016
 */
public class ParamDatabase implements Iterable<Param<?>>, Serializable, XMLSerializable {

    private static final long serialVersionUID = -194866493580603428L;

    /**
     * The database is in a Map.
     */
    private transient FastMap<String, Param<?>> theDB = new FastMap<>();

    //-----------------------------------------------------------------------------
    /**
     * Default constructor to return a reference to an empty parameter database.
     */
    public ParamDatabase() {
    }

    //-----------------------------------------------------------------------------
    /**
     * Returns the number of elements contained in the database.
     *
     * @return The number of elements contained in the database.
     */
    public int size() {
        return theDB.size();
    }

    /**
     * Returns true if the database contains no elements.
     *
     * @return Returns true if the database contains no elements, false otherwise.
     */
    public boolean isEmpty() {
        return theDB.isEmpty();
    }

    /**
     * Returns the key set for this parameter database. This is the set of parameter
     * labels stored in the database.
     *
     * @return The key set for this parameter database.
     */
    public Set<String> keySet() {
        return theDB.keySet();
    }

    /**
     * Returns true if the database contains an element for the given label.
     *
     * @param label The label for the parameter that we are looking for. May not be null.
     * @return Returns true if the database contains an element for the given label.
     */
    public boolean containsKey(String label) {
        return theDB.containsKey(requireNonNull(label));
    }

    /**
     * Gets the parameter associated with the specified label in the database.
     *
     * @param label The specified parameter label. May not be null.
     * @return The parameter for the label or null if the label is not defined in the
     *         database.
     * @see ParamDatabase#put
     */
    public Param<?> get(String label) {
        return theDB.get(requireNonNull(label));
    }

    /**
     * Puts the specified parameter into the database, using the specified label. The
     * parameter may be retrieved by doing a get() with the same label. The label and the
     * parameter cannot be null.
     *
     * @param label The specified label in the database. May not be null.
     * @param value The parameter to be added to the database. May not be null.
     * @return The old value of the label, or null if it did not exist in the database.
     * @see ParamDatabase#get
     */
    public Param<?> put(String label, Param<?> value) {
        return theDB.put(requireNonNull(label), requireNonNull(value));
    }

    /**
     * Puts all the parameters in the specified parameter database into this database.
     *
     * @param pdb The database to have all the parameter values added to this one. May not
     *            be null.
     */
    public void putAll(ParamDatabase pdb) {
        theDB.putAll(pdb.theDB);
    }

    /**
     * Removes the parameter corresponding to the label. Does nothing if the label is not
     * present.
     *
     * @param label The label that needs to be removed. May not be null.
     * @return The value of label, or null if the label was not found.
     */
    public Param<?> remove(String label) {
        return theDB.remove(requireNonNull(label));
    }

    /**
     * Clears the database so that it has no more elements in it.
     */
    public void clear() {
        theDB.clear();
    }

    /**
     * Returns an iteration of the elements.
     */
    @Override
    public Iterator<Param<?>> iterator() {
        return theDB.values().iterator();
    }

    /**
     * Return the entry set for this database.
     *
     * @return The entry set for this database.
     */
    public Set<Entry<String, Param<?>>> entrySet() {
        return theDB.entrySet();
    }

    /**
     * Converts database to a rather lengthy String.
     *
     * @return String representation of the entire database.
     */
    @Override
    public String toString() {
        return theDB.toString();
    }

    /**
     * Method that lists out the contents of this database to a print stream; one
     * parameter per line.
     *
     * @param out The print stream to list the parameters out to. May not be null.
     */
    public void list(PrintStream out) {
        requireNonNull(out);
        for (Iterator<Param<?>> i = iterator(); i.hasNext();) {
            Param<?> aParam = i.next();
            out.println(aParam.toString());
        }
    }

    //-----------------------------------------------------------------------------
    /**
     * During serialization, this will write out each parameter in this parameter
     * database.
     */
    private void writeObject(java.io.ObjectOutputStream out) throws IOException {

        // Call the default write object method.
        out.defaultWriteObject();

        // Write out the number of methods we are going to write out.
        int size = theDB.size();
        out.writeInt(size);

        // Now loop over all the methods and write out each one.
        for (Param<?> theParam : this) {
            out.writeObject(theParam);
        }

    }

    /**
     * During de-serialization, this will handle the reading in of the list of methods
     * associated with this parameter.
     */
    private void readObject(java.io.ObjectInputStream in)
            throws IOException, ClassNotFoundException {

        // Call the default read object method.
        in.defaultReadObject();

        // Create a new database.
        theDB = new FastMap<>();

        // Read in the number of methods in the file.
        int size = in.readInt();

        // Loop over each method and read it in.
        for (int i = 0; i < size; ++i) {
            Param<?> theParam = (Param<?>)in.readObject();

            // Add the new method to the method list.
            theDB.put(theParam.getLabel(), theParam);
        }

    }

    /**
     * Holds the default XML representation for this object.
     */
    protected static final XMLFormat<ParamDatabase> XML = new XMLFormat<ParamDatabase>(ParamDatabase.class) {
        @Override
        public void read(XMLFormat.InputElement xml, ParamDatabase obj) throws XMLStreamException {
            while (xml.hasNext()) {
                Param<?> param = xml.getNext();
                obj.put(param.getLabel(), param);
            }
        }

        @Override
        public void write(ParamDatabase obj, XMLFormat.OutputElement xml) throws XMLStreamException {
            Set<Map.Entry<String, Param<?>>> entries = obj.entrySet();
            for (Entry<String, Param<?>> entry : entries) {
                xml.add(entry.getValue());
            }
        }
    };

}
