/*
*   ARfn_b_Cr_Ct - Method for calculating AR based on b, Cr, and Ct.
*
*   Copyright (C) 1999-2023 by Joseph A. Huwaldt
*   All rights reserved.
*   
*   This library is free software; you can redistribute it and/or
*   modify it under the terms of the GNU Lesser General Public
*   License as published by the Free Software Foundation; either
*   version 2.1 of the License, or (at your option) any later version.
*   
*   This library is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
*   Library General Public License for more details.
*
*  You should have received a copy of the GNU Lesser General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*  Or visit:  http://www.gnu.org/licenses/lgpl.html
**/
package jahuwaldt.js.pmethods.winggeom;

import javax.measure.quantity.Dimensionless;
import javax.measure.unit.SI;

import jahuwaldt.js.param.Parameter;
import jahuwaldt.js.ParamTools.*;


/**
*  <p>  Calculates planform wing aspect ratio based on planform span,
*       wing root chord length, and wing tip chord length.
*  </p>
*
*  <p>  Modified by:  Joseph A. Huwaldt   </p>
*
*  @author   Joseph A. Huwaldt   Date:  February 6, 1999
*  @version  December 16, 2023
**/
public class ARfn_b_Cr_Ct extends AbstractParamMethod<Dimensionless> {

	/**
	*  Constructor for this method.
	*
	*  @param  b   Label for the "span" parameter.
	*  @param  Cr  Label for the "root chord" parameter.
	*  @param  Ct  Label for the "tip chord" parameter.
	**/
	public ARfn_b_Cr_Ct( String b, String Cr, String Ct ) {
		_inputLabels = new String [ 3 ];
		_inputLabels[0] = b;
		_inputLabels[1] = Cr;
		_inputLabels[2] = Ct;
	}

	/**
	*  Calculates planform wing aspect ratio based on planform
	*  wing span,  wing root chord length, and wing tip chord
	*  length.
	*  Method:  AR = 2*b/(Cr + Ct)
	*
	*  @param  inputs  Array of input parameters.
	*                  [0=planform span (b), 1=wing root chord length (Cr),
	*                   2=wing tip chord length (Ct)].
	*  @return  Planform aspect ratio (AR).
	**/
    @Override
	public Parameter<Dimensionless> calc( Parameter[] inputs ) {
		double b = inputs[0].doubleValue(SI.METER);
		double Cr = inputs[1].doubleValue(SI.METER);
		double Ct = inputs[2].doubleValue(SI.METER);
		double AR = 2 * b / (Cr + Ct);
		return Parameter.valueOf(AR, Dimensionless.UNIT);
	}


}


