/*
 *   Crfn_taper_Ct - Method for calculating Cr as a function of taper and Ct.
 *
 *   Copyright (C) 1999-2016, by Joseph A. Huwaldt.
 *   All rights reserved.
 *   
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/lgpl.html
 */
package jahuwaldt.js.pmethods.winggeom;

import jahuwaldt.js.ParamTools.*;
import jahuwaldt.js.param.Parameter;
import javax.measure.quantity.Length;

/**
 * Calculates wing root chord length based on wing taper ratio, and tip chord length.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt, Date: February 6, 1999
 * @version March 26, 2016
 */
@SuppressWarnings("serial")
public class Crfn_taper_Ct extends AbstractParamMethod<Length> {

    /**
     * Constructor for this method.
     *
     * @param taper Label for the "taper ratio" parameter.
     * @param Ct    Label for the "tip chord" parameter.
     */
    public Crfn_taper_Ct(String taper, String Ct) {
        _inputLabels = new String[2];
        _inputLabels[0] = taper;
        _inputLabels[1] = Ct;
    }

    /**
     * Calculates wing root chord length based on wing taper ratio, and tip chord length.
     * <p>
     * Method:  <code>Cr = Ct/taper</code>
     * </p>
     *
     * @param inputs Array of input parameters. [0=wing taper ratio (taper), 1=tip chord
     *               length (Ct)].
     * @return Root chord length (Cr) in base units of length.
     */
    @Override
    public Parameter<Length> calc(Parameter[] inputs) {
        Parameter taper = inputs[0];
        Parameter Ct = inputs[1];
        Parameter Cr = Ct.divide(taper);		//	Cr = Ct / taper
        return Cr;
    }

}
