/*
 *   Ctfn_S_b_Cr - Method for calculating Ct as a function of S, b and Cr.
 *
 *   Copyright (C) 1998-2016, by Joseph A. Huwaldt.
 *   All rights reserved.
 *   
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/lgpl.html
 */
package jahuwaldt.js.pmethods.winggeom;

import jahuwaldt.js.ParamTools.*;
import jahuwaldt.js.param.Parameter;
import javax.measure.quantity.Length;

/**
 * Calculates wing tip chord length based on planform area, planform span and root chord
 * length.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt, Date: February 6, 1999
 * @version March 26, 2016
 */
@SuppressWarnings("serial")
public class Ctfn_S_b_Cr extends AbstractParamMethod<Length> {

    /**
     * Constructor for this method.
     *
     * @param S  Label for the "planform area" parameter.
     * @param b  Label for the "span" parameter.
     * @param Cr Label for the "root chord" parameter.
     */
    public Ctfn_S_b_Cr(String S, String b, String Cr) {
        _inputLabels = new String[3];
        _inputLabels[0] = S;
        _inputLabels[1] = b;
        _inputLabels[2] = Cr;
    }

    /**
     * Calculates wing tip chord length based on planform area, planform span and root
     * chord length.
     * <p>
     * Method:  <code>Ct = 2*S/b - Cr</code>
     * </p>
     *
     * @param inputs Array of input parameters. [0=planform area (S), 1=planform span (b),
     *               2=root chord length(Cr)].
     * @return Wing tip chord length (Ct) in base units of length.
     */
    @Override
    public Parameter<Length> calc(Parameter[] inputs) {
        Parameter S = inputs[0];
        Parameter b = inputs[1];
        Parameter Cr = inputs[2];
        Parameter Ct = S.times(2).divide(b).minus(Cr);
        return Ct.asType(Length.class);
    }

}
