/*
 *   Sfn_b_Cr_Ct - Method for calculating S as a function of b, Cr, and Ct.
 *
 *   Copyright (C) 1998-2016, by Joseph A. Huwaldt.
 *   All rights reserved.
 *   
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/lgpl.html
 */
package jahuwaldt.js.pmethods.winggeom;

import jahuwaldt.js.ParamTools.*;
import jahuwaldt.js.param.Parameter;
import javax.measure.quantity.Area;

/**
 * Calculates planform wing area based on planform span, wing root chord length and wing
 * tip chord length.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt, Date: February 6, 1999
 * @version March 26, 2016
 */
@SuppressWarnings("serial")
public class Sfn_b_Cr_Ct extends AbstractParamMethod<Area> {

    /**
     * Constructor for this method.
     *
     * @param b  Label for the "span" parameter.
     * @param Cr Label for the "root chord" parameter.
     * @param Ct Label for the "tip chord" parameter.
     */
    public Sfn_b_Cr_Ct(String b, String Cr, String Ct) {
        _inputLabels = new String[3];
        _inputLabels[0] = b;
        _inputLabels[1] = Cr;
        _inputLabels[2] = Ct;
    }

    /**
     * Calculates planform wing area based on planform wing span, wing root chord length
     * and wing tip chord length.
     * <p>
     * Method: <code>S = b/2*(Cr + Ct)</code>
     * </p>
     *
     * @param inputs Array of input parameters. [0=planform span (b), 1=wing root chord
     *               length (Cr), 2=wing tip chord length (Ct)].
     * @return	Planform wing area (S) in base units of area.
     */
    @Override
    public Parameter<Area> calc(Parameter[] inputs) {
        Parameter b = inputs[0];
        Parameter Cr = inputs[1];
        Parameter Ct = inputs[2];
        Parameter S = b.divide(2).times(Cr.plus(Ct));		//	S = b * 0.5 * (Cr + Ct);
        return S.asType(Area.class);
    }

}
