/*
*   bfn_yMAC_etaMAC - Method for calculating b as a function of yMac and etaMAC.
*
*   Copyright (C) 1998-2023 by Joseph A. Huwaldt
*   All rights reserved.
*   
*   This library is free software; you can redistribute it and/or
*   modify it under the terms of the GNU Lesser General Public
*   License as published by the Free Software Foundation; either
*   version 2.1 of the License, or (at your option) any later version.
*   
*   This library is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
*   Library General Public License for more details.
*
*  You should have received a copy of the GNU Lesser General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*  Or visit:  http://www.gnu.org/licenses/lgpl.html
**/
package jahuwaldt.js.pmethods.winggeom;


import jahuwaldt.js.ParamTools.AbstractParamMethod;
import jahuwaldt.js.param.Parameter;
import javax.measure.quantity.Length;


/**
*  <p>  Calculates planform wing span based on spanwise position of MAC (Cbar),
*       and fractional spanwise position of MAC.
*  </p>
*
*  <p>  Modified by:  Joseph A. Huwaldt  </p>
*
*  @author   Joseph A. Huwaldt   Date:  February 6, 1999
*  @version  December 16, 2023
**/
public class bfn_yMAC_etaMAC extends AbstractParamMethod<Length> {

	/**
	*  Constructor for this method.
	*
	*  @param  yMAC    Label for the "spanwise location of MAC" parameter.
	*  @param  etaMAC  Label for the "% spanwise location of MAC" parameter.
	**/
	public bfn_yMAC_etaMAC( String yMAC, String etaMAC ) {
		_inputLabels = new String [ 2 ];
		_inputLabels[0] = yMAC;
		_inputLabels[1] = etaMAC;
	}

	/**
	*  Calculates planform wing span based on spanwise position of MAC (Cbar),
	*  and fractional spanwise position of MAC.
	*  Method:  b = 2*yMAC/etaMAC
	*
	*  @param  inputs  Array of input parameters.
	*                  [0=spanwise location of MAC (yMAC),
	*                   1=fractional spanwise location of MAC (etaMAC)].
	*  @return  Planform wing span (b) in base length units.
	**/
    @Override
	public Parameter<Length> calc( Parameter[] inputs ) {
		Parameter yMAC = inputs[0];
		Parameter etaMAC = inputs[1];
		Parameter b = yMAC.times(2).divide(etaMAC);		//	b = 2 * yMAC / etaMAC
		return b.asType(Length.class);
	}


}


