/*
 *   etaMACfn_taper - Method for calculating etaMAC as a function of taper.
 *
 *   Copyright (C) 1998-2016, by Joseph A. Huwaldt.
 *   All rights reserved.
 *   
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/lgpl.html
 */
package jahuwaldt.js.pmethods.winggeom;

import jahuwaldt.js.ParamTools.*;
import jahuwaldt.js.param.Parameter;
import javax.measure.quantity.Dimensionless;

/**
 * Calculates fractional spanwise location of MAC (Cbar) based on wing taper ratio.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt, Date: February 6, 1999
 * @version March 26, 2016
 */
@SuppressWarnings("serial")
public class etaMACfn_taper extends AbstractParamMethod<Dimensionless> {

    /**
     * Constructor for this method.
     *
     * @param taper Label for the "taper ratio" parameter.
     */
    public etaMACfn_taper(String taper) {
        _inputLabels = new String[1];
        _inputLabels[0] = taper;
    }

    /**
     * Calculates fractional spanwise location of MAC (Cbar) based on wing taper ratio.
     * <p>
     * Method: <code>etaMAC = 1/3*(1 + 2*taper)/(1 + taper)</code>
     * </p>
     *
     * @param inputs Array of input parameters. [0=wing taper ratio (taper)].
     * @return Spanwise location of the MAC (Cbar) as a fraction of span (b).
     */
    @Override
    public Parameter<Dimensionless> calc(Parameter[] inputs) {
        double taper = inputs[0].getValue();
        double etaMAC = (1 + 2 * taper) / (1 + taper) / 3.0;
        return Parameter.valueOf(etaMAC, Dimensionless.UNIT);
    }

}
