/*
 *   taperfn_swpLE_swp4_AR - Method for calculating taper as a function of sweepLE, sweep4 and AR.
 *
 *   Copyright (C) 1998-2016, by Joseph A. Huwaldt.
 *   All rights reserved.
 *   
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/lgpl.html
 */
package jahuwaldt.js.pmethods.winggeom;

import jahuwaldt.js.ParamTools.*;
import jahuwaldt.js.param.Parameter;
import javax.measure.quantity.Dimensionless;
import javax.measure.unit.SI;

/**
 * Calculates wing taper ratio based on leading edge sweep angle, quarter chord sweep
 * angle and wing aspect ratio.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt, Date: February 6, 1999
 * @version March 26, 2016
 */
@SuppressWarnings("serial")
public class taperfn_swpLE_swp4_AR extends AbstractParamMethod<Dimensionless> {

    /**
     * Constructor for this method.
     *
     * @param sweepLE Label for the "leading edge sweep" parameter.
     * @param sweep4  Label for the "1/4 chord sweep" parameter.
     * @param AR      Label for the "aspect ratio" parameter.
     */
    public taperfn_swpLE_swp4_AR(String sweepLE, String sweep4, String AR) {
        _inputLabels = new String[3];
        _inputLabels[0] = sweepLE;
        _inputLabels[1] = sweep4;
        _inputLabels[2] = AR;
    }

    /**
     * Calculates wing taper ratio based on leading edge sweep angle, quarter chord sweep
     * angle and wing aspect ratio.
     * <p>
     * Method:
     * <code>taper = (1-AR*(tan(sweepLE)-tan(sweep4)))/(1+AR*(tan(sweepLE)-tan(sweep4)))</code>
     * </p>
     *
     * @param inputs Array of input parameters. [0=leading edge sweep angle (sweepLE),
     *               1=quarter chord sweep angle (sweep4) 2=wing aspect ratio (AR)].
     * @return Wing taper ratio (taper).
     */
    @Override
    public Parameter<Dimensionless> calc(Parameter[] inputs) {
        double sweepLE = inputs[0].doubleValue(SI.RADIAN);
        double sweep4 = inputs[1].doubleValue(SI.RADIAN);
        double AR = inputs[2].getValue();
        double temp = Math.tan(sweepLE) - Math.tan(sweep4);
        double taper = (1 - AR * temp) / (1 + AR * temp);
        return Parameter.valueOf(taper, Dimensionless.UNIT);
    }

}
