/*
 *   xMACfn_xMAC4_Cbar - Method for calculating xMAC as a function of xMAC/4 and Cbar.
 *
 *   Copyright (C) 2011-2016, by Joseph A. Huwaldt.
 *   All rights reserved.
 *   
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/lgpl.html
 */
package jahuwaldt.js.pmethods.winggeom;

import jahuwaldt.js.ParamTools.*;
import jahuwaldt.js.param.Parameter;
import javax.measure.quantity.Length;

/**
 * Calculates longitudinal location of leading edge of the MAC (Cbar) from the wing apex
 * based on longitudinal location of MAC/4 from the apex and MAC length.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt, Date: September 11, 2011
 * @version March 26, 2016
 */
@SuppressWarnings("serial")
public class xMACfn_xMAC4_Cbar extends AbstractParamMethod<Length> {

    /**
     * Constructor for this method.
     *
     * @param xMAC4 Label for the "longitudinal location of MAC/4" parameter.
     * @param Cbar  Label for the "MAC length" parameter.
     */
    public xMACfn_xMAC4_Cbar(String xMAC4, String Cbar) {
        _inputLabels = new String[2];
        _inputLabels[0] = xMAC4;
        _inputLabels[1] = Cbar;
    }

    /**
     * Calculates longitudinal location of MAC (Cbar) leading edge from the wing apex
     * based on spanwise location of MAC/4 and MAC length.
     * <p>
     * Method: <code>xMAC = xMAC4 - Cbar/4</code>
     * </p>
     *
     * @param inputs Array of input parameters. [0=longitudinal locatin of MAC/4 (xMAC4),
     *               1=MAC length (Cbar)].
     * @return Longitudinal location of the MAC (Cbar) leading edge from wing apex in base
     *         units of length.
     */
    @Override
    public Parameter<Length> calc(Parameter[] inputs) {
        Parameter xMAC4 = inputs[0];
        Parameter Cbar = inputs[1];
        Parameter xMAC = xMAC4.minus(Cbar.divide(4));
        return xMAC;
    }

}
